/*
 * Copyright (c) 2011 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 ** morCntrlTaskCodels.c
 **
 ** Codels used by the control task morCntrlTask
 **
 **/

#include <math.h>
#include <stdio.h>
#include <portLib.h>

#include "server/morHeader.h"
#include "morConst.h"

/*------------------------------------------------------------------------
 * recalConfig  -  control codel of CONTROL request SetConfig
 *
 * Description:    
 * Report: OK
 *                 S_mor_TRACKING_ON
 *
 * Returns:    OK or ERROR
 */

STATUS
recalConfig(GENPOS_CART_CONFIG *setPosStr, int *report)
{
	GENPOS_CART_STATE *configId;

	configId = &morDataStrId->robot;

	/* Do not allow position change if the robot is moving */
	if (morDataStrId->mode = MOR_TRACK) {
		*report = S_mor_TRACKING_ON;
		return ERROR;
	}
	configId->xRob = setPosStr->x;
	configId->yRob = setPosStr->y;
	configId->theta = setPosStr->theta;
	return OK;
}

/*------------------------------------------------------------------------
 * controlRef  -  control codel of CONTROL request SetRef
 *
 * Description:    
 * Report: OK
 *                 S_mor_BAD_REF_DATA
 *
 * Returns:    OK or ERROR
 */

STATUS
controlRef(GENPOS_CART_REF *ref, int *report)
{
	GENPOS_CART_REF *refId = &SDI_F->ref;
	MOR_MODE mode  = SDI_F->mode;
	int retval = OK;

	/* Check if mode changed */
	switch (mode) {
	case MOR_EMERGENCY:
		*report = S_mor_EMERGENCY_STOP;
		return ERROR;
	case MOR_MOTORS_OFF:
		*report = S_mor_MOTORS_OFF;
		return ERROR;
	}
	/* Start correct tracking mode */
	*report = 0;
	switch (ref->dataType) {
	case GENPOS_POS_AND_SPEED_DATA:
		mode = MOR_TRACK;
		morDataStrId->trackStr.trackMode = GENPOS_TRACK_CONFIG;
		break;
	case GENPOS_SPEED_DATA:
		mode = MOR_TRACK;
		morDataStrId->trackStr.trackMode = GENPOS_TRACK_SPEED;
		break;
	default:
		*report = S_mor_BAD_REF;
		return ERROR;
	}
	memcpy(refId, ref, sizeof(GENPOS_CART_REF));
	return OK;
}

/*------------------------------------------------------------------------
 * controlCmdParam  -  control codel of CONTROL request SetCmdParam
 *
 * Description:    
 * Report: OK
 *                 S_mor_BAD_CMD_PARAMETERS
 *
 * Returns:    OK or ERROR
 */

STATUS
controlCmdParam(GENPOS_CART_CMD_PARAM *cmd, int *report)
{
  /* ... add your code here ... */
  return OK;
}

/*------------------------------------------------------------------------
 * computStopRef  -  control codel of CONTROL request Stop
 *
 * Description:    
 * Report: OK
 *
 * Returns:    OK or ERROR
 */

STATUS
computStopRef(int *report)
{
	GENPOS_CART_STATE *robotId;           /* Config robot */
	GENPOS_CART_REF *refId;               /* Consigne */
	double dtArret, dtRetard;  /* Temps prise en compte et temps deceleration */
	double dSArret, dTArret;   /* Longueurs lin et ang de decel */
	double angleFinal;         /* Angle a l'arrivee */
	double period;

	robotId = &morDataStrId->robot;
	refId = &morDataStrId->ref;

	/* Longueur d'arret */
	period = EXEC_TASK_PERIOD(MOR_MOTIONTASK_NUM);
	dtArret = MAX (fabs(robotId->v) / MOR_ACCEL_LIN_MAX,
		       fabs(robotId->w) / MOR_ACCEL_ANG_MAX);
	dtRetard =  1 * period;
	dSArret = robotId->v * (dtArret / 2 + dtRetard);
	dTArret = robotId->w * (dtArret / 2 + dtRetard);
	angleFinal = robotId->theta + dTArret;

	/* Consigne */
	refId->x = robotId->xRob + dSArret * cos (angleFinal);
	refId->y = robotId->yRob + dSArret * sin (angleFinal);
	refId->v = 0. ;
	refId->vt = 0.;
	refId->w = 0. ;
	return(OK);
}

/*------------------------------------------------------------------------
 * joystickCtrl  -  control codel of EXEC request JoystickOn
 *
 * Description:    
 * Report: OK
 *                 S_mor_BAD_REF
 *                 S_mor_MOR_LIB_ERROR
 *                 S_mor_JOYSTICK_ERROR
 *                 S_mor_MOTORS_OFF
 *                 S_mor_EMERGENCY_STOP
 *
 * Returns:    OK or ERROR
 */

STATUS
joystickCtrl(int *report)
{
	int retval = OK;

	/* Check if base initialized */
	if (!SDI_F->initParams.initBase) {
		*report = S_mor_BASE_NOT_INITIALIZED;
		return ERROR;
	}
	/* Check if mode changed */
	switch (SDI_F->mode) {
	case MOR_EMERGENCY:
		*report = S_mor_EMERGENCY_STOP;
		return ERROR;
	case MOR_MOTORS_OFF:
		*report = S_mor_MOTORS_OFF;
		return ERROR;
	}
	return OK;
}

/*------------------------------------------------------------------------
 * trackStart  -  control codel of EXEC request Track
 *
 * Description:    
 * Report: OK
 *                 S_mor_POSTER_NOT_FOUND
 *                 S_mor_BAD_REF
 *                 S_mor_CMD_STOP_TRACK
 *
 * Returns:    OK or ERROR
 */

STATUS
trackStart(MOR_TRACK_PARAMS *params, int *report)
{
	int retval = OK;

	/* Check if base initialized */
	if (!SDI_F->initParams.initBase) {
		*report = S_mor_BASE_NOT_INITIALIZED;
		return ERROR;
	}
	/* Check if mode changed */
	switch (morDataStrId->mode) {
	case MOR_EMERGENCY:
		*report = S_mor_EMERGENCY_STOP;
		return ERROR;
	case MOR_MOTORS_OFF:
		*report = S_mor_MOTORS_OFF;
		return ERROR;
	}
	/* Start correct tracking mode */
	*report = 0;
	switch (params->mode) {
	case GENPOS_NO_TRACKING:
		morDataStrId->mode = MOR_READY;
		morDataStrId->trackStr.trackMode = params->mode;
		*report = S_mor_CMD_STOP_TRACK;
		retval = ERROR;
		break;
	case GENPOS_TRACK_POS:
	case GENPOS_TRACK_CONFIG:
	case GENPOS_TRACK_SPEED:
		morDataStrId->trackStr.trackMode = params->mode;
		break;
	default:
		*report = S_mor_BAD_REF;
		retval = ERROR;
	}
	return retval;
}


/*------------------------------------------------------------------------
 * moveSpineCntrl  -  control codel of EXEC request MoveSpine
 *
 * Description:    
 * Report: OK
 *                 S_mor_SPINE_NOT_INITIALIZED
 *                 S_mor_MOTORS_OFF
 *                 S_mor_EMERGENCY_STOP
 *
 * Returns:    OK or ERROR
 */

STATUS
moveSpineCntrl(MOR_SPINE_REF *spineRef, int *report)
{

	if (!SDI_F->initParams.initSpine) {
		*report = S_mor_SPINE_NOT_INITIALIZED;
		return ERROR;
	}
	if (spineRef->pos < MOR_SPINE_POS_MIN ||
	    spineRef->pos > MOR_SPINE_POS_MAX ||
	    fabs(spineRef->speed) < MOR_SPINE_SPEED_MIN ||
	    fabs(spineRef->speed) > MOR_SPINE_SPEED_MAX) {
		*report = S_mor_BAD_REF;
		return ERROR;
	}
	return OK;
}
