/*
 * Copyright (c) 2011 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 ** morMotionTaskCodels.c
 **
 ** Codels called by execution task morMotionTask
 **
 **/
#include <sys/types.h>
#include <errno.h>
#include <math.h>
#include <stdbool.h>

#include <portLib.h>
#include <h2mathLib.h>

#include "server/morHeader.h"

#include <mor/pscan.h>
#include <mor/harmonica.h>
#include <mor/relayboard.h>
#include <mor/morLib.h>

#include "morConst.h"
#include "server/morHeader.h"

#include "joystick.h"

static MOR_STR *mor = NULL;
static POSTER_ID pomPosterId = NULL;
static JOYSTICK_STR *joy = NULL;

static STATUS initPomPosters(void);
static STATUS updatePomPosters(void);
static void getJoystickSpeeds(double *, double *, double *, double *, double *);
/*------------------------------------------------------------------------
 *
 * initOdoAndAsserv  --  Initialization codel (fIDS, ...)
 *
 * Description:
 *
 * Returns:    OK or ERROR
 */

STATUS
initOdoAndAsserv(int *report)
{
	GENPOS_CART_STATE *configId;
	GENPOS_CART_REF *refId;
	GENPOS_CART_CMD_PARAM *cmdId;
	MOR_SPINE_REF *spineRef;
	MOR_SPINE_STATE *spineState;

	morDataStrId->mode = MOR_UNINITIALIZED;

	/* configuration */
	configId = &morDataStrId->robot;
	configId->xRob = 0.;
	configId->yRob = 0.;
	configId->theta = 0.;
	configId->xRef = MOR_DIST_TO_CONTROLLED_POINT;
	configId->yRef = 0;
	configId->v = 0;
	configId->w = 0;

	/* Ref Data */
	refId = &morDataStrId->ref;
	refId->dataType = GENPOS_POS_AND_SPEED_DATA;
	refId->backFlag = GENPOS_FORWARD_MOTION;
	refId->x = configId->xRef;
	refId->y = configId->yRef;
	refId->theta = configId->theta;
	refId->v = 0.;
	refId->w = 0.;

	refId->vmax = MOR_VMAX;
	refId->wmax = MOR_WMAX;
	refId->linAccelMax = MOR_ACCEL_LIN_MAX;
	refId->angAccelMax = MOR_ACCEL_ANG_MAX;

	/* Control parameters */
	cmdId = &morDataStrId->cmd;
	cmdId->distPoint = MOR_DIST_TO_CONTROLLED_POINT;
	cmdId->KpLongit = MOR_KP_LONGIT;
	cmdId->KiLongit = MOR_KI_LONGIT;
	cmdId->KpTransv = MOR_KP_TRANSV;
	cmdId->KiTransv = MOR_KI_TRANSV;

	cmdId->KpAng = MOR_KP_ANG;
	cmdId->KiAng = MOR_KI_ANG;

	cmdId->longitMaxGap = MOR_LONGIT_MAX_GAP;
	cmdId->transvMaxGap = MOR_TRANSV_MAX_GAP;
	cmdId->angMaxGap    = MOR_ANG_MAX_GAP;

	cmdId->longitErrorMax = MOR_LONGIT_ERROR_MAX;
	cmdId->transvErrorMax = MOR_TRANSV_ERROR_MAX;
	cmdId->angErrorMax = MOR_ANG_ERROR_MAX;

	cmdId->longitMaxErrSum = MOR_LONGIT_MAX_ERR_SUM;
	cmdId->transvMaxErrSum = MOR_TRANSV_MAX_ERR_SUM;
	cmdId->angMaxErrSum = MOR_ANG_MAX_ERR_SUM;

	/* spine */
	spineRef = &morDataStrId->spineRef;
	spineRef->speed = 0.;

	spineState = &morDataStrId->spineState;
	spineState->speed = 0.;

	/* Initialize Pom */
	initPomPosters();

	return OK;
}

/*------------------------------------------------------------------------
 *
 * stopOdoAndAsserv  --  Termination codel
 *
 * Description:
 *
 * Returns:    OK or ERROR
 */

STATUS
stopOdoAndAsserv(void)
{
	if (mor == NULL)
		return OK;

	if (mor->relayboard)
		relayboardUpdate(mor->relayboard,
	    "                                                            ");
	if (mor == NULL)
		return OK;
	morStop(mor);
	morEnd(mor);
	printf("bye\n");
	return OK;
}

/*------------------------------------------------------------------------
 *
 * odo  --  1st codel of permanent activity (called BEFORE other activities)
 *
 * Description:
 *
 * Reports:      OK
 *              S_mor_EMERGENCY_STOP
 *
 * Returns:    OK or ERROR
 */

static
void
copyRawData(void)
{
	MOR_ROBOT_DATA *data = &morDataStrId->rawData;
	CAN_HARMONICA_STR *steer, *drive;
	MOR_WHEEL_DATA *wd;
	int i;

	data->relayState = mor->relayboard->state;
	data->charge_c = mor->relayboard->charge_c;
	data->voltage = mor->relayboard->voltage;
	data->temperature = mor->relayboard->temperature;
	for (i = 0; i < MOR_N_WHEELS; i++) {
		steer = mor->wheels[i].steer;
		drive = mor->wheels[i].drive;
		wd = &data->wheels[i];
		harmonicaGetPosVelRadS(steer, &wd->steerPos, &wd->steerVel);
		harmonicaGetPosVelRadS(drive, &wd->drivePos, &wd->driveVel);
		harmonicaGetCurrent(steer, &wd->steerCurrent);
		harmonicaGetCurrent(drive, &wd->driveCurrent);
		wd->steerMotorState = harmonicaGetMotorState(steer);
		wd->driveMotorState = harmonicaGetMotorState(drive);
	}
}

STATUS
odo(int *report)
{
	GENPOS_CART_STATE *robot = &SDI_F->robot;
	MOR_ODO_STR *odo;
	MOR_SPINE_STATE *spine;

	if (mor == NULL || SDI_F->mode == MOR_UNINITIALIZED)
		return OK;

	odo = mor->odo;
	if (mor->initBase) {
		copyRawData();
		morOdo(mor, EXEC_TASK_PERIOD(MOR_MOTIONTASK_NUM));
		robot->xRob = odo->x;
		robot->xRef = odo->x;
		robot->yRob = odo->y;
		robot->yRef = odo->y;
		robot->theta = odo->theta;
		robot->v = odo->vx;
		robot->vt = odo->vy;
		robot->w = odo->w;
		if (mor->motorFailure) {
			if (SDI_F->mode != MOR_MOTORS_OFF) {
				printf("Mor: motor failure!\n");
				SDI_F->mode = MOR_MOTORS_OFF;
			}
		} else {
			if (SDI_F->mode == MOR_MOTORS_OFF) {
				printf("Mor: motor recovered\n");
				SDI_F->mode = MOR_READY;
			}
		}
	}
	if (mor->relayboard->state != 0) {
		if (SDI_F->mode != MOR_EMERGENCY) {
			printf("Mor: Emergency stop!\n");
			SDI_F->mode = MOR_EMERGENCY;
		}
	} else {
		if (SDI_F->mode == MOR_EMERGENCY) {
			printf("Mor: Emergency stop cleared\n");
			SDI_F->mode = MOR_MOTORS_OFF;
		}
	}

	/* spine */
	if (!mor->initSpine)
		return OK;
	spine = &morDataStrId->spineState;
	harmonicaGetPosVelRadS(mor->spine, &spine->pos, &spine->speed);
	spine->stopSwitches = 0;
	if (mor->spine->digIn & HARMONICA_INPUT_FORWARD_LIMIT)
		spine->stopSwitches |= MOR_SPINE_STOP_HIGH;
	if (mor->spine->digIn & HARMONICA_INPUT_REVERSE_LIMIT)
		spine->stopSwitches |= MOR_SPINE_STOP_LOW;

	return OK;
}
/*----------------------------------------------------------------------*/

/**
 ** Actual tracking code
 **/
static void
track(int *report, double *vRef, double *vtRef, double *wRef)
{
	GENPOS_CART_REF *refId = &SDI_F->ref;
	GENPOS_CART_STATE *robot = &SDI_F->robot;
	GENPOS_TRACK *trackId = &SDI_F->trackStr;
	static double prevV, prevVt, prevW;
	static BOOL prevDrifted = FALSE;
	BOOL drifted = FALSE;

	switch (trackId->trackMode) {
	case GENPOS_TRACK_POS:
		if (refId->dataType != GENPOS_POS_DATA) {
			fprintf(stderr, "wrong data type for TRACK_POS\n");
			*report = S_mor_BAD_REF;
			return;
		}
		drifted = contLawCartPositionControl(&morDataStrId->cmd,
		    refId, robot, EXEC_TASK_PERIOD(MOR_MOTIONTASK_NUM),
		    prevV, prevW, vRef, wRef);
		*vtRef = 0.;
		break;

	case GENPOS_TRACK_CONFIG:
		if (refId->dataType != GENPOS_POS_AND_SPEED_DATA) {
			fprintf(stderr, "wrong data type for TRACK_CONFIG\n");
			*report = S_mor_BAD_REF;
			return;
		}
		drifted = contLawCartConfigControl(&morDataStrId->cmd,
		    refId, robot, EXEC_TASK_PERIOD(MOR_MOTIONTASK_NUM),
		    prevV, prevW, vRef, wRef);
		*vtRef = 0.;
		break;

	case GENPOS_TRACK_SPEED:
		if (refId->dataType != GENPOS_SPEED_DATA) {
			fprintf(stderr, "wrong data type for TRACK_SPEED %d\n",
			    refId->dataType);
			*report = S_mor_BAD_REF;
			return;
		}
		*vRef = (refId->backFlag == GENPOS_FORWARD_MOTION ? refId->v :
		    -refId->v);
		*vtRef = refId->vt;
		*wRef = refId->w;
		break;
	default:
		*report = S_mor_BAD_REF;
		break;
	}
	if (drifted && !prevDrifted) {
		printf("odoAndAsserv: important drift!\n");
		printf("ref: %.2lf %.2lf %.2lf\n",
		       refId->x, refId->y, refId->theta);
		printf("pos: %.2lf %.2lf %.2lf\n",
		       robot->xRef, robot->yRef, robot->theta);
	}
	if (!drifted && prevDrifted)
		printf("odoAndAsserv: drift cancelled\n");
	prevDrifted = drifted;

	prevV = *vRef;
	prevVt = *vtRef;
	prevW = *wRef;
}

/*------------------------------------------------------------------------
 *
 * asserv  --  2nd codel of permanent activity (called AFTER other activities)
 *
 * Description:
 *
 * Reports:      OK
 *              S_mor_EMERGENCY_STOP
 *
 * Returns:    OK or ERROR
 */

STATUS
asserv(int *report)
{
	GENPOS_CART_REF *refId = &SDI_F->ref;
	GENPOS_CART_STATE *robot = &SDI_F->robot;
	double vxRef, vyRef, wRef;

	*report = OK;
	if (mor == NULL)
		return OK;	/* not initialized yet */

	switch (SDI_F->mode) {
	case MOR_UNINITIALIZED:
	case MOR_MOTORS_OFF:
	case MOR_EMERGENCY:
		/* no motion possible */
		return OK;
	case MOR_READY:
		vxRef = 0.;
		vyRef = 0.;
		wRef = 0.;
		refId->x = robot->xRef;
		refId->y = robot->yRef;
		refId->theta = robot->theta;
		break;
	case MOR_MANUAL:
		if (joystickPoll(joy) < 0) {
			*report = S_mor_JOYSTICK_ERROR;
			break;
		}
		refId->x = robot->xRef;
		refId->y = robot->yRef;
		refId->theta = robot->theta;
		getJoystickSpeeds(&vxRef, &vyRef, &wRef,
		    &refId->linAccelMax, &refId->angAccelMax);
		refId->v = vxRef;
		refId->vt = vyRef;
		refId->w = wRef;
		break;
	case MOR_TRACK:
		track(report, &vxRef, &vyRef, &wRef);
		break;
	default:
		printf("mode %d\n", SDI_F->mode);
		*report = S_mor_BAD_REF;
		return ERROR;
	}
	if (*report != OK) {
		/* In case an error occured,
		   stop the robot and the tracking */
		SDI_F->mode = MOR_READY;
		SDI_F->trackStr.trackMode = GENPOS_NO_TRACKING;
		vxRef = 0.;
		vyRef = 0.;
		wRef = 0.;
	}

	/* Compute commands to motors */
	morMove(mor, vxRef, vyRef, wRef,
	    EXEC_TASK_PERIOD(MOR_MOTIONTASK_NUM));
	/* Send them */
	morSendSpeed(mor, EXEC_TASK_PERIOD(MOR_MOTIONTASK_NUM));
	/* Query digital inputs of spine axis */
	if (mor->initSpine)
		harmonicaRequestDigIn(mor->spine);
	return OK;
}

/*------------------------------------------------------------------------
 * Init
 *
 * Description:
 *
 * Reports:      OK
 *              S_mor_MORLIB_ERROR
 *              S_mor_MOTORS_OFF
 *              S_mor_EMERGENCY_STOP
 */

/* morInitStart  -  codel START of Init
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
morInitStart(MOR_INIT_PARAMS *initParams, int *report)
{
	int i;

	if (mor != NULL) {
		printf("stopping running control\n");
		morStop(mor);
		morEnd(mor);
	}
	mor = morInit(initParams->canDev.name,
	    initParams->relayBoardDev.name,
	    initParams->initBase,
	    initParams->initSpine);
	if (mor == NULL) {
		*report = S_mor_MORLIB_ERROR;
		return ETHER;
	}
	if (morInitPlatform(mor) == ERROR) {
		*report = S_mor_MOTORS_OFF;
		return ETHER;
	}
	/* XXX homing should be cut into codels and put in initMain... */
	if (morHoming(mor) == ERROR)
		*report = S_mor_EMERGENCY_STOP;
	/* Start watchdogs */
	if (initParams->initSpine)
		harmonicaStartWatchdog(mor->spine, true);
	for (i = 0; i < 4; i++) {
		harmonicaStartWatchdog(mor->wheels[i].steer, true);
		harmonicaStartWatchdog(mor->wheels[i].drive, true);
	}
	/* start read task */
	if (morStartReadTask(mor) != OK) {
		*report = S_mor_MORLIB_ERROR;
		return ETHER;
	}
	if (mor->relayboard != NULL && morStartRelayboardTask(mor) != OK) {
		*report = S_mor_MORLIB_ERROR;
		return ETHER;
	}
	return EXEC;
}

/* morInitMain  -  codel EXEC of Init
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
morInitMain(MOR_INIT_PARAMS *initParams, int *report)
{
	int i;
	bool ok = true;
	double pos, dpos, svel, dvel;

	*report = OK;

	if (!initParams->initBase)
		return END;

	if (mor->motorFailure) {
		*report = S_mor_MOTORS_OFF;
		return FAIL;
	}
	morMove(mor, 0., 0., 0., EXEC_TASK_PERIOD(MOR_MOTIONTASK_NUM));
	morSendSpeed(mor, EXEC_TASK_PERIOD(MOR_MOTIONTASK_NUM));

	/* Wait for motors to reach initial position */
	for (i = 0; i < MOR_N_WHEELS; i++) {
		harmonicaGetDeltaPosVelRadS(mor->wheels[i].steer,
		    &dpos, &svel);
		harmonicaGetDeltaPosVelRadS(mor->wheels[i].drive,
		    &dpos, &dvel);
		if (fabs(svel) > 0.01 || fabs(dvel) > 0.01)
			ok = false;
	}
	if (!ok)
		return EXEC;
	return END;
}

/* morInitEnd  -  codel END of Init
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
morInitEnd(MOR_INIT_PARAMS *initParams, int *report)
{
	int i;

	if (initOdoAndAsserv(report) == ERROR)
		return FAIL;
	SDI_F->mode = MOR_READY;
	printf("Robot ready\n");
	return ETHER;
}

/*------------------------------------------------------------------------
 * JoystickOn
 *
 * Description:
 *
 * Reports:      OK
 *              S_mor_BAD_REF
 *              S_mor_RMP_LIB_ERROR
 *              S_mor_JOYSTICK_ERROR
 *              S_mor_MOTORS_OFF
 *              S_mor_EMERGENCY_STOP
 */

/* morJoystickOnStart  -  codel START of JoystickOn
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
morJoystickOnStart(int *report)
{
	int n;

	joy = joystickInit(MOR_JOYSTICK_DEVICE_PATH);
	if (joy == NULL) {
		fprintf(stderr, "joystickInit(%s) failed: %s\n",
			MOR_JOYSTICK_DEVICE_PATH, strerror(errno));
		*report = S_mor_JOYSTICK_ERROR;
		return ETHER;
	}
	n = joystickPoll(joy);
	memset(joy->buttons, 0, sizeof(joy->buttons));
	SDI_F->mode = MOR_MANUAL;
	return EXEC;
}

/* morJoystickOnMain  -  codel EXEC of JoystickOn
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
morJoystickOnMain(int *report)
{

	/* Check if mode changed */
	switch (SDI_F->mode) {
	case MOR_EMERGENCY:
		*report = S_mor_EMERGENCY_STOP;
		return END;
	case MOR_MOTORS_OFF:
		*report = S_mor_MOTORS_OFF;
		return END;
	}
	/* Check for user abort (button 4) */
	if (joy->buttons[3]) {
		printf("Stop joystick\n");
		return morJoystickOnInter(report);
	}
	return EXEC;
}

/* morJoystickOnInter  -  codel INTER of JoystickOn
   Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
morJoystickOnInter(int *report)
{
	GENPOS_CART_REF *refId;
	int n;

	n = joystickPoll(joy);

	refId = &morDataStrId->ref;
	refId->linAccelMax = MOR_ACCEL_LIN_MAX;
	refId->angAccelMax = MOR_ACCEL_ANG_MAX;
	SDI_F->mode = MOR_READY;
	joystickEnd(joy);
	joy = NULL;
	return ETHER;
}


/*------------------------------------------------------------------------
 * SetPom
 *
 * Description:
 *
 * Reports:      OK
 *              S_mor_POSTER_NOT_FOUND
 */

/* setPom  -  codel EXEC of SetPom
   Returns:  EXEC END ETHER FAIL ZOMBIE */

ACTIVITY_EVENT
setPom(GENPOS_POSTER_NAME *posterName, int *report)
{
	POSTER_ID id;

	if (posterFind(posterName->name, &id) != OK) {
		logMsg("posterFind(%s) failed\n", posterName->name);
		*report = S_mor_POSTER_NOT_FOUND;
	} else
		pomPosterId = id;
	return ETHER;
}

/*----------------------------------------------------------------------*/
/*
 * Initialize POM
 */
static STATUS
initPomPosters(void)
{
	POM_ME_POS *mePos = &morDataStrId->mePos;

	mePos->kind = POM_ME_ABSOLUTE;
	mePos->date1 = -1;
	mePos->confidence = 0.2;	// XXX used?
	mePos->main.var.voo     = MOR_VAR_TT;
	mePos->main.var.vpp     = MOR_VAR_TT;
	mePos->main.var.vrr     = MOR_VAR_TT;
	mePos->main.var.vxx     = MOR_VAR_XX;
	mePos->main.var.vyy     = MOR_VAR_YY;
	mePos->main.var.vzz     = MOR_VAR_XX;

	return OK;
}

/*----------------------------------------------------------------------*/

/*
 * Update POM poster
 */
static STATUS
updatePomPosters(void)
{
	POM_ME_POS *mePos = &SDI_F->mePos;
	GENPOS_CART_STATE *configId;
	POM_POS pomPos;

	if (pomPosterId != NULL) {
		if(morPOM_POSPosterRead(pomPosterId, &pomPos) == OK) {
			mePos->date1 = pomPos.date;
		} else {
			logMsg("morPOM_POSPosterRead() failed\n");
			pomPosterId = NULL;
			mePos->date1 = -1;
		}
	} else
		mePos->date1 = -1;

	configId = &morDataStrId->robot;
	mePos->main.euler.yaw = configId->theta;
	mePos->main.euler.x   = configId->xRob;
	mePos->main.euler.y   = configId->yRob;

	mePos->main.euler.pitch = 0.;
	mePos->main.euler.roll = 0.;
	mePos->main.euler.z = 0.;

	/* speeds are mostly zero, just update the linear
	   and angular components */
	mePos->VLocal.euler.x   = configId->v;
	mePos->VLocal.euler.y   = configId->vt;
	mePos->VLocal.euler.yaw = configId->w;

	/* all variances and co-variances remain zero */

	return OK;
}

/*----------------------------------------------------------------------*/

/*
 * Convert joystick values to reference speed
 *
 * A square function is used to get more precision at low speeds.
 *
 */
static void
getJoystickSpeeds(double *vx, double *vy, double *w,
    double *avmax, double *awmax)
{
	double vgainw = -2, vgainv = -2, againw = -2, againv = -2;
	if (joy->buttons[0]) {
		vgainw = 0.7; vgainv = 0.4; againw = 0.4; againv = 0.2;
	}
	if (joy->buttons[1]) {
		vgainw = 1.0; vgainv = 1.0; againw = 0.5; againv = 0.4;
	}
	if (joy->buttons[2]) {
		vgainw = 1.3; vgainv = 2.0; againw = 0.6; againv = 0.6;
	}
	if (vgainv > -1) {
		*vx = -SIGN(joy->axes[1]) * joy->axes[1] * joy->axes[1]
		  * MOR_JOYSTICK_VGAIN * vgainv;
		*vy = 0.;
		*w = -SIGN(joy->axes[0]) * joy->axes[0] * joy->axes[0]
		  * MOR_JOYSTICK_WGAIN * vgainw;
		if (avmax) *avmax = MOR_ACCEL_LIN_MAX * againv;
		if (awmax) *awmax = MOR_ACCEL_ANG_MAX * againw;
	} else {
		*vx = 0.;
		*vy = 0.;
		*w = 0.;
	}
}

/*------------------------------------------------------------------------
 * MoveSpine
 *
 * Description:
 *
 * Reports:      OK
 *              S_mor_SPINE_NOT_INITIALIZED
 *              S_mor_MOTORS_OFF
 *              S_mor_EMERGENCY_STOP
 */

static double maxSpeed = 0.0;
static int spineDirection;

/* moveSpineStart  -  codel START of MoveSpine
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
moveSpineStart(MOR_SPINE_REF *ref, int *report)
{
	maxSpeed = 0.0;
	if (SDI_F->spineState.pos < ref->pos)
		spineDirection = 1;
	else
		spineDirection = -1;

	harmonicaSetGearPosVelRadS(mor->spine, ref->pos, ref->speed);
	return EXEC;
}

/* moveSpineMain  -  codel EXEC of MoveSpine
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
moveSpineMain(MOR_SPINE_REF *ref, int *report)
{
	double pos, speed;
	MOR_SPINE_STATE *spineState = &SDI_F->spineState;
	MOR_SPINE_REF *spineRef = &SDI_F->spineRef;

	harmonicaGetPosVelRadS(mor->spine, &pos, &speed);
	if (fabs(speed) > maxSpeed)
		maxSpeed = fabs(speed);
	if (spineState->stopSwitches & MOR_SPINE_STOP_LOW &&
	    spineDirection < 0) {
		printf("reached low switch\n");
		harmonicaSetGearVelRadS(mor->spine, 0.);
		return ETHER;
	}
	if (spineState->stopSwitches & MOR_SPINE_STOP_HIGH &&
	    spineDirection > 0) {
		printf("reached high switch\n");
		harmonicaSetGearVelRadS(mor->spine, 0.);
		return ETHER;
	}
	if (fabs(SDI_F->spineState.pos - ref->pos) > 0.001)
		return EXEC;
	harmonicaSetGearVelRadS(mor->spine, 0.);
	printf("max speed: %.4lf\n", maxSpeed);
	return ETHER;
}

/* moveSpineInter  -  codel INTER of MoveSpine
   Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
moveSpineInter(MOR_SPINE_REF *ref, int *report)
{
	/* Stop where we are */
	harmonicaSetGearVelRadS(mor->spine, 0.);
	printf("max speed: %.4lf\n", maxSpeed);
	return ETHER;
}
